import os
import yaml
from flask_babel import Babel
from flask import Flask, request, session, redirect, url_for
from .blog import blog, db_loader
from .seo import seo
from .plugin_loader import plugify
from flask_minify import Minify

from flaskext.markdown import Markdown


def create_app(config_path):
    app = Flask(__name__)
    Markdown(app)
    # print(os.getcwd())
    absolute_config_path = os.path.join(os.getcwd(), config_path)

    app.config.from_file(absolute_config_path, load=yaml.safe_load)
    app.config["CONFIG_PATH"] = absolute_config_path
    db_driver = db_loader.load_db_driver(app.config["DB"]["type"])
    app.db = db_driver.get_db(app.config)
    app.babel = Babel(app)

    blog_blueprint = blog.create_blog_blueprint(db=app.db,
                                                config=app.config,
                                                babel=app.babel,
                                                url_prefix="/")

    seo_blueprint = seo.create_seo_blueprint(db=app.db,
                                             config=app.config,
                                             url_prefix="/")
    app.register_blueprint(blog_blueprint)
    app.register_blueprint(seo_blueprint)

    Minify(app=app, html=True, js=True, cssless=True)



    @app.babel.localeselector
    def get_locale():
        return session.get('language', request.accept_languages.best_match(app.config["LANG_MAP"].keys()))

    @app.route('/language/<language>')
    def set_language(language):
        session['language'] = language
        return redirect(url_for('index'))
    return plugify(app)



import os
import sys
from os.path import dirname, abspath
from importlib.util import spec_from_file_location, module_from_spec


def load_db_driver(db_type):
    db_dir = os.path.join(dirname(dirname(abspath(__file__))), 'db')
    spec = spec_from_file_location(db_type, os.path.join(db_dir, db_type + ".py"))
    db_driver = module_from_spec(spec)
    sys.modules[db_type] = db_driver
    spec.loader.exec_module(db_driver)
    return db_driver



#TODO rename file, extract it to another library, remove qgl and aiohttp from dependencies

from gql import gql, Client
from gql.transport.aiohttp import AIOHTTPTransport
import json

from platzky.blog.db import DB


def get_db(config):
    endpoint = config["DB"]["CMS_ENDPOINT"]
    token = config["DB"]["CMS_TOKEN"]
    return GraphQL(endpoint, token)


class GraphQL(DB):
    def __init__(self, endpoint, token):
        full_token = 'bearer ' + token
        transport = AIOHTTPTransport(url=endpoint, headers={'Authorization': full_token})
        self.client = Client(transport=transport)

    def get_all_posts(self, lang):
        all_posts = gql(
            """
            query MyQuery($lang: Lang!) {
              posts(where: {language: $lang},  orderBy: date_DESC, stage: PUBLISHED){
                createdAt
                date
                title
                excerpt
                slug
                tags
                coverImage
                {
                    url
                }                
              }
            }
            """
        )
        return self.client.execute(all_posts, variable_values={"lang": lang})['posts']

    def get_post(self, slug):
        post = gql(
            """
            query MyQuery($slug: String!) {
              post(where: {slug: $slug}, stage: PUBLISHED) {
                title
                contentInMarkdown
                excerpt
                tags
                coverImage
                {
                    url
                }
                comments {
                    author
                    comment
                    date: createdAt
                }                
              }
            }
            """)
        return self.client.execute(post, variable_values={"slug": slug})['post']

    #TODO Cleanup page logic of internationalization (now it depends on translation of slugs)
    def get_page(self, slug):
        post = gql(
            """
            query MyQuery ($slug: String!){
              page(where: {slug: $slug}, stage: PUBLISHED) {
                title
                contentInMarkdown
                coverImage
                {
                    url
                }
              }
            }
            """)
        return self.client.execute(post, variable_values={"slug": slug})['page']

    def get_posts_by_tag(self, tag, lang):
        post = gql(
            """
            query MyQuery ($tag: String!, $lang: Lang!){
              posts(where: {tags_contains_some: [$tag], language: $lang}, stage: PUBLISHED) {
                    tags
                    title
                    slug
                    excerpt
                    date
                    coverImage
                    {
                        url
                    }
              }
            }
            """)
        return self.client.execute(post, variable_values={"tag": tag, "lang": lang})['posts']

    def get_all_providers(self):
        all_providers = gql(
            """
            query MyQuery {
              providers(stage: PUBLISHED) {
                link
                name
                offer
                currency
              }
            }
            """
        )
        providers = self.client.execute(all_providers)["providers"]
        for provider in providers:
            provider["offer"] = json.loads(provider["offer"])
        return providers

    def get_all_questions(self):
        all_questions = gql(
            """
            query MyQuery {
              questions(stage: PUBLISHED) {
                question
                field
                inputType
              }
            }
            """
        )
        query = self.client.execute(all_questions)
        return query['questions']

    def get_redirections(self):
        redirections = gql(
            """
            query MyQuery{
              redirections(stage: PUBLISHED){
                source
                destination
              }
            }
            """
        )
        return {x['source']:x['destination'] for x in self.client.execute(redirections)['redirections']}

    def add_comment(self, author_name, comment, post_slug):
        add_comment = gql(
            """
            mutation MyMutation($author: String!, $comment: String!, $slug: String!) {
                createComment(
                    data: {
                        author: $author,
                        comment: $comment,
                        post: {connect: {slug: $slug}}
                    }
                ) {
                    id
                }
            }
            """)
        self.client.execute(add_comment, variable_values={
            "author": author_name, "comment": comment, "slug": post_slug
        })



from flask import redirect
from functools import partial


def process(app):
    redirects = app.db.get_redirections()
    for source, destiny in redirects.items():
        func = partial(redirect, destiny, code=301)
        func.__name__ = f"{source}-{destiny}"
        app.route(rule=source)(func)
    return app



import urllib.parse
from flask import request, render_template, make_response, Blueprint


def create_seo_blueprint(db, config, url_prefix):
    seo = Blueprint('seo', __name__, url_prefix=url_prefix)
    languages = config["LANG_MAP"]
    main_domain = config["MAIN_DOMAIN"]
    # secure_headers = SecureHeaders()

    @seo.route("/robots.txt")
    def robots():
        robots_response = render_template("robots.txt", domain=request.host, mimetype='text/plain')
        response = make_response(robots_response)
        response.headers["Content-Type"] = "text/plain"
        return response

    @seo.route("/sitemap.xml", host=languages['pl']['domain'])
    def pl_sitemap():
        return sitemap('pl')

    @seo.route("/sitemap.xml", host=main_domain)
    def main_sitemap():
        return sitemap('en')

    def sitemap(lang):
        """
        Route to dynamically generate a sitemap of your website/application.
        lastmod and priority tags omitted on static pages.
        lastmod included on dynamic content such as seo posts.
        """

        global url
        host_components = urllib.parse.urlparse(request.host_url)
        host_base = host_components.scheme + "://" + host_components.netloc

        # Static routes with static content
        static_urls = list()
        for rule in seo.url_map.iter_rules():
            if not str(rule).startswith("/admin") and not str(rule).startswith("/user"):
                if "GET" in rule.methods and len(rule.arguments) == 0:
                    url = {
                        "loc": f"{host_base}{str(rule)}"
                    }
                static_urls.append(url)

        # Dynamic routes with dynamic content
        dynamic_urls = list()
        seo_posts = db.get_all_posts(lang)
        for post in seo_posts:
            slug = post['slug']
            datet = post['date'].split('T')[0]
            url = {
                "loc": f"{host_base}/seo/{slug}",
                "lastmod": datet
            }
            dynamic_urls.append(url)

        statics = list({v['loc']: v for v in static_urls}.values())
        dynamics = list({v['loc']: v for v in dynamic_urls}.values())
        xml_sitemap = render_template("sitemap.xml", static_urls=statics, dynamic_urls=dynamics,
                                      host_base=host_base)
        response = make_response(xml_sitemap)
        response.headers["Content-Type"] = "application/xml"
        return response

    return seo



from flask_wtf import FlaskForm
from wtforms import StringField, SubmitField
from wtforms.validators import DataRequired
from wtforms.widgets import TextArea
from flask_babel import lazy_gettext


class CommentForm(FlaskForm):
    author_name = StringField(lazy_gettext('Name'), validators=[DataRequired()])
    comment = StringField(lazy_gettext('Type comment here'), validators=[DataRequired()], widget=TextArea())
    submit = SubmitField(lazy_gettext('Comment'))



import urllib.parse
from flask import request, render_template, current_app, redirect, send_from_directory, make_response, url_for, Blueprint


#from secure import SecureHeaders

from platzky.blog import comment_form, post_formatter
from .www_handler import redirect_www_to_nonwww, redirect_nonwww_to_www



def create_blog_blueprint(db, config, babel, url_prefix):
    blog = Blueprint('blog', __name__, url_prefix=url_prefix)
    languages = config["LANG_MAP"]
    main_domain = config["MAIN_DOMAIN"]
    # secure_headers = SecureHeaders()

    @blog.before_request
    def handle_www_redirection():
        if config["USE_WWW"]:
            return redirect_nonwww_to_www()
        else:
            return redirect_www_to_nonwww()


    # @blog.after_request
    # def set_secure_headers(response):
    #     secure_headers.flask(response)
    #     return response

    def domains_locale(dom):
        return config["DOMAIN_TO_LANG"][dom]

    @blog.errorhandler(404)
    def page_not_found(e):
        return render_template('404.html', 404)

    @babel.localeselector
    def get_locale():
        return domains_locale(request.host)

    @blog.context_processor
    def utils():
        return {'languages': languages,
                "language": get_locale(),
                "url_link": lambda x: urllib.parse.quote(x, safe='')
                }

    @blog.route('/', methods=["GET"])
    def index():
        lang = get_locale()
        return render_template("blog.html", posts=db.get_all_posts(lang))

    @blog.route('/feed', methods=["GET"])
    def get_feed():
        lang = get_locale()

        response = make_response(render_template("feed.xml", posts=db.get_all_posts(lang)))
        response.headers["Content-Type"] = "application/xml"
        return response

    @blog.route('/<post_slug>', methods=["GET", "POST"])
    def get_post(post_slug):
        if request.method == "POST":
            comment = request.form.to_dict()
            db.add_comment(post_slug=post_slug, author_name=comment["author_name"],
                                            comment=comment["comment"])
            return redirect(url_for('get_post', post_slug=post_slug, comment_sent=True))
        raw_post = db.get_post(post_slug)

        return render_template("post.html", post=post_formatter.format_post(raw_post), post_slug=post_slug,
                               form=comment_form.CommentForm(), comment_sent=request.args.get('comment_sent'))

    @blog.route('/page/<path:page_slug>', methods=["GET", "POST"])
    def get_page(page_slug):
        return render_template("page.html", post=db.get_page(page_slug))


    @blog.route('/tag/<path:tag>', methods=["GET"])
    def get_posts_from_tag(tag):
        lang = get_locale()
        posts = db.get_posts_by_tag(tag, lang)
        return render_template("blog.html", posts=posts, subtitle=f" - tag: {tag}")

    @blog.route('/lang/<string:lang>', methods=["GET"])
    def change_language(lang):
        new_domain = blog.get_langs_domain(lang)
        return redirect("http://" + new_domain, code=301)

    @blog.route('/results', methods=["POST"])
    def render_answers():
        answers = {}
        required_fields = current_app.config['fields']
        for field, value in request.form.to_dict().items():
            if field not in required_fields:
                continue
            answers[field] = value
        return redirect(url_for('get_results', **answers))

    @blog.route('/icon/<string:name>', methods=["GET"])
    def icon(name):
        return send_from_directory('../static/icons', f"{name}.png")

    return blog

















#
#
# class MainApp(Flask):
#     def __init__(self, config):
#         super().__init__(__name__, static_folder=None)
#         self.url_map.host_matching = True
#         self.config.update(config)
#         self.config["BABEL_TRANSLATION_DIRECTORIES"] = os.path.join(os.path.dirname(os.path.realpath(__file__)),
#                                                                     "../locale") #TODO create poetry plugin and let it handle this
#         app_engine_init()
#
#         self.babel = Babel(self)
#         # self.mail = Mail()
#         self.db = get_db(config["DB"])
#         questions = self.db.get_all_questions()
#         self.questions = questioner.get_questions(questions, config["fields"])
#         self.language_map = config["LANG_MAP"]
#
#         self.accepted_languages = self.language_map.keys()
#         self.providers = self.db.get_all_providers()
#         Markdown(self)
#
#     def does_lang_has_domain(self, lang):
#         return 'domain' in self.language_map.get(lang)
#
#     def get_langs_domain(self, lang):
#         if self.does_lang_has_domain(lang):
#             return self.language_map.get(lang)['domain']
#         else:
#             return self.config['MAIN_DOMAIN']
#
#     def domains_locale(self, dom):
#         return self.config["DOMAIN_TO_LANG"][dom]
#
#     def multi_route(self, *args, **kwargs):
#         def insider(func):
#             for domain in self.config["DOMAINS"]:
#                 self.add_url_rule(*args, endpoint=func.__name__, view_func=func, **kwargs, host=domain)
#             return func
#
#         return insider



import datetime
import humanize


def get_delta(now, date_in):
    date = datetime.datetime.strptime(date_in.split(".")[0], "%Y-%m-%dT%H:%M:%S")
    delta = humanize.naturaltime(now - date)
    return delta


def format_post(post):
    now = datetime.datetime.now()
    for comment in post["comments"]:
        comment.update({"time_delta": get_delta(now, comment["date"])})
    post["comments"].sort(key=lambda x: x["date"], reverse=True)
    return post



import os
import sys
from os.path import dirname, abspath
import importlib.util


def find_plugins():
    plugin_dir = os.path.join(dirname(abspath(__file__)), 'plugins')
    plugins = []
    for name in os.listdir(plugin_dir):
        if "plugin" in name:
            module_name = name.removesuffix('.py')
            spec = importlib.util.spec_from_file_location(module_name,
                                                          os.path.join(plugin_dir, name))
            plugin = importlib.util.module_from_spec(spec)
            sys.modules[module_name] = plugin
            spec.loader.exec_module(plugin)
            plugins.append(plugin)
    return plugins


def plugify(app):
    for plugin in find_plugins():
        plugin.process(app)
    return app



from platzky.blog.db import DB
import json
import datetime
import os.path

def get_db(config):
    db_path = os.path.join(os.path.dirname(config["CONFIG_PATH"]),
                           config["DB"]["PATH"])
    return JsonFile(db_path)


class JsonFile(DB):
    def __init__(self, file_path):
        self.file = file_path
        with open(self.file) as json_file:
            self.json = json.load(json_file)

    def _save_file(self):
        with open(self.file, 'w') as json_file:
            json.dump(self.json, json_file)

    def get_all_posts(self, lang):
        return self.json["posts"]

    def get_post(self, slug):
        post = next(filter(lambda x: x["slug"] == slug, self.json["posts"]))
        return post

    def get_page(self, slug, lang):
        post = next(filter(lambda x: x["slug"] == slug, self.json["pages"]))
        return post

    def get_posts_by_tag(self, tag, lang):
        posts = filter(lambda x: tag in x["tags"], self.json["posts"])
        return posts

    def get_all_providers(self):
        return self.json["providers"]

    def get_all_questions(self):
        return self.json["questions"]

    def get_redirections(self):
        return self.json["redirections"]

    def add_comment(self, author_name, comment, post_slug):
        comment = {
            "author": str(author_name),
            "comment": str(comment),
            "date": datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
        }
        post_index = next(i for i in range(len(self.json["posts"])) if self.json["posts"][i]["slug"] == post_slug)
        self.json["posts"][post_index]["comments"].append(comment)
        self._save_file()



import urllib.parse
from flask import request, redirect


def redirect_nonwww_to_www():
    """Redirect non-www requests to www."""
    urlparts = urllib.parse.urlparse(request.url)
    if not urlparts.netloc.startswith("www."):
        urlparts = urlparts._replace(netloc=f'www.{urlparts.netloc}')
        return redirect(urllib.parse.urlunparse(urlparts), code=301)


def redirect_www_to_nonwww():
    """Redirect www requests to non-www."""
    urlparts = urllib.parse.urlparse(request.url)
    if urlparts.netloc.startswith("www."):
        urlparts = urlparts._replace(netloc=urlparts.netloc.removeprefix("www."))
        return redirect(urllib.parse.urlunparse(urlparts), code=301)



class DB:
    def get_all_posts(self, lang):
        pass

    def get_post(self, slug):
        pass

    def get_page(self, slug):
        pass

    def get_posts_by_tag(self, tag, lang):
        pass

    def add_comment(self, author_name, comment, post_slug):
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['platzky',
 'platzky.blog',
 'platzky.db',
 'platzky.plugins.redirections',
 'platzky.seo']

package_data = \
{'': ['*'], 'platzky': ['static/*', 'templates/*']}

install_requires = \
['Flask-Babel>=2.0.0,<3.0.0',
 'Flask-Markdown>=0.3,<0.4',
 'Flask-Minify>=0.39,<0.40',
 'Flask-WTF>=1.0.1,<2.0.0',
 'Flask>=2.2.2,<3.0.0',
 'PyYAML>=6.0,<7.0',
 'aiohttp>=3.8.3,<4.0.0',
 'gql>=3.4.0,<4.0.0',
 'humanize>=4.3.0,<5.0.0']

setup_kwargs = {
    'name': 'platzky',
    'version': '0.1.0',
    'description': 'Another blog in python',
    'long_description': '# platzky\n\nBlog engine in python \n',
    'author': 'Your Name',
    'author_email': 'you@example.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
